import { resolve, relative } from 'pathe';
import { defu } from 'defu';
import { joinURL } from 'ufo';
import { pascalCase, camelCase } from 'scule';
import { useNuxt, defineNuxtModule, useLogger, createResolver, tryResolveModule, addServerHandler, addImportsSources, addTemplate } from '@nuxt/kit';

async function generateDeclarationTypes(endpoints, globalOpenAPIOptions) {
  const resolvedSchemaEntries = await Promise.all(
    Object.entries(endpoints).filter(([, endpoint]) => Boolean(endpoint.schema)).map(async ([id, endpoint]) => {
      const types = await generateSchemaTypes({ id, endpoint, openAPITSOptions: globalOpenAPIOptions });
      return [id, types];
    })
  );
  return resolvedSchemaEntries.map(
    ([id, types]) => `
declare module '#nuxt-api-party/${id}' {
  ${types.replace(/^/gm, "  ").trimEnd()}
}`.trimStart()
  ).join("\n\n").trimStart();
}
async function generateSchemaTypes(options) {
  const { default: openAPITS, astToString } = await import('openapi-typescript');
  const schema = await resolveSchema(options.endpoint);
  try {
    const ast = await openAPITS(schema, {
      ...options.openAPITSOptions,
      ...options.endpoint.openAPITS
    });
    return astToString(ast);
  } catch (error) {
    console.error(`Failed to generate types for ${options.id}`);
    console.error(error);
    return `
export type paths = Record<string, never>
export type webhooks = Record<string, never>
export interface components {
  schemas: never
  responses: never
  parameters: never
  requestBodies: never
  headers: never
  pathItems: never
}
export type $defs = Record<string, never>
export type operations = Record<string, never>
`.trimStart();
  }
}
async function resolveSchema({ schema }) {
  const nuxt = useNuxt();
  if (typeof schema === "function")
    return await schema();
  if (typeof schema === "string" && !isValidUrl(schema))
    return new URL(resolve(nuxt.options.rootDir, schema), import.meta.url);
  return schema;
}
function isValidUrl(url) {
  try {
    return Boolean(new URL(url));
  } catch (e) {
    return false;
  }
}

const module = defineNuxtModule({
  meta: {
    name: "nuxt-api-party",
    configKey: "apiParty",
    compatibility: {
      nuxt: "^3.7"
    }
  },
  defaults: {
    endpoints: {},
    client: false,
    openAPITS: {},
    server: {
      basePath: "__api_party"
    }
  },
  async setup(options, nuxt) {
    const moduleName = "nuxt-api-party";
    const logger = useLogger(moduleName);
    const getRawComposableName = (endpointId) => `$${camelCase(endpointId)}`;
    const getDataComposableName = (endpointId) => `use${pascalCase(endpointId)}Data`;
    if (!Object.keys(options.endpoints).length && !nuxt.options.runtimeConfig.apiParty)
      logger.error("Missing any API endpoint configuration. Please check the `apiParty` module configuration in `nuxt.config.ts`.");
    nuxt.options.runtimeConfig.apiParty = defu(
      nuxt.options.runtimeConfig.apiParty,
      options
    );
    if (!nuxt.options.ssr) {
      logger.info("Enabling Nuxt API Party client requests by default because SSR is disabled.");
      options.client = "always";
    }
    const resolvedOptions = nuxt.options.runtimeConfig.apiParty;
    nuxt.options.runtimeConfig.public.apiParty = defu(
      nuxt.options.runtimeConfig.public.apiParty,
      resolvedOptions.client ? resolvedOptions : {
        // Only expose cookies endpoint option to the client
        endpoints: Object.fromEntries(
          Object.entries(resolvedOptions.endpoints).map(
            ([endpointId, endpoint]) => [endpointId, { cookies: endpoint.cookies }]
          )
        ),
        client: false,
        server: resolvedOptions.server
      }
    );
    const { resolve } = createResolver(import.meta.url);
    nuxt.options.build.transpile.push(resolve("runtime"));
    const relativeTo = (path) => relative(
      resolve(nuxt.options.rootDir, nuxt.options.buildDir, "module"),
      resolve(path)
    );
    const endpointKeys = Object.keys(resolvedOptions.endpoints);
    const schemaEndpoints = Object.fromEntries(
      Object.entries(resolvedOptions.endpoints).filter(([, endpoint]) => "schema" in endpoint)
    );
    const schemaEndpointIds = Object.keys(schemaEndpoints);
    const hasOpenAPIPkg = await tryResolveModule("openapi-typescript", [nuxt.options.rootDir]);
    if (schemaEndpointIds.length && !hasOpenAPIPkg) {
      logger.warn("OpenAPI types generation is enabled, but the `openapi-typescript` package is not found. Please install it to enable endpoint types generation.");
      schemaEndpointIds.length = 0;
    }
    addServerHandler({
      route: joinURL("/api", options.server.basePath, ":endpointId"),
      method: "post",
      handler: resolve("runtime/server/handler")
    });
    nuxt.hook("nitro:config", (config) => {
      config.externals ||= {};
      config.externals.inline ||= [];
      config.externals.inline.push(...[
        resolve("runtime/utils"),
        resolve("runtime/formData"),
        resolve("runtime/server/$api")
      ]);
      config.alias ||= {};
      config.alias[`#${moduleName}/server`] = resolve(nuxt.options.buildDir, `module/${moduleName}-nitro`);
      config.virtual ||= {};
      config.virtual[`#${moduleName}/server`] = () => `
import { _$api } from '${resolve("runtime/server/$api")}'
${endpointKeys.map((i) => `
export const ${getRawComposableName(i)} = (...args) => _$api('${i}', ...args)
`.trimStart()).join("")}`.trimStart();
      if (schemaEndpointIds.length) {
        config.typescript ||= {};
        config.typescript.tsConfig ||= {};
        config.typescript.tsConfig.include ||= [];
        config.typescript.tsConfig.include.push(`./module/${moduleName}-schema.d.ts`);
      }
      config.imports = defu(config.imports, {
        presets: [{
          from: `#${moduleName}/server`,
          imports: endpointKeys.map(getRawComposableName)
        }]
      });
    });
    addImportsSources({
      from: resolve(nuxt.options.buildDir, `module/${moduleName}`),
      imports: endpointKeys.flatMap((i) => [getRawComposableName(i), getDataComposableName(i)])
    });
    nuxt.options.alias[`#${moduleName}`] = resolve(nuxt.options.buildDir, `module/${moduleName}`);
    addTemplate({
      filename: `module/${moduleName}.mjs`,
      getContents() {
        return `
import { _$api } from '${resolve("runtime/composables/$api")}'
import { _useApiData } from '${resolve("runtime/composables/useApiData")}'
${endpointKeys.map((i) => `
export const ${getRawComposableName(i)} = (...args) => _$api('${i}', ...args)
export const ${getDataComposableName(i)} = (...args) => _useApiData('${i}', ...args)
`.trimStart()).join("")}`.trimStart();
      }
    });
    addTemplate({
      filename: `module/${moduleName}.d.ts`,
      getContents() {
        return `
// Generated by ${moduleName}
import type { ApiClient, OpenAPIClient, ApiClientFetchOptions, OpenAPIClientFetchOptions } from '${relativeTo("runtime/composables/$api")}'
import type { UseApiData, UseOpenAPIData, UseApiDataOptions, UseOpenAPIDataOptions } from '${relativeTo("runtime/composables/useApiData")}'

${schemaEndpointIds.map((i) => `
import type { paths as ${pascalCase(i)}Paths } from '#${moduleName}/${i}'
`.trimStart()).join("").trimEnd()}

export type { ApiClient, OpenAPIClient, ApiClientFetchOptions, OpenAPIClientFetchOptions, UseApiData, UseOpenAPIData, UseApiDataOptions, UseOpenAPIDataOptions }

${endpointKeys.map((i) => `
export declare const ${getRawComposableName(i)}: ${schemaEndpointIds.includes(i) ? `OpenAPIClient<${pascalCase(i)}Paths>` : "ApiClient"}
export declare const ${getDataComposableName(i)}: ${schemaEndpointIds.includes(i) ? `UseOpenAPIData<${pascalCase(i)}Paths>` : "UseApiData"}
`.trimStart()).join("").trimEnd()}
`.trimStart();
      }
    });
    addTemplate({
      filename: `module/${moduleName}-nitro.d.ts`,
      async getContents() {
        return `
// Generated by ${moduleName}
export { ${endpointKeys.map(getRawComposableName).join(", ")} } from './${moduleName}'
`.trimStart();
      }
    });
    if (schemaEndpointIds.length) {
      addTemplate({
        filename: `module/${moduleName}-schema.d.ts`,
        async getContents() {
          return await generateDeclarationTypes(schemaEndpoints, resolvedOptions.openAPITS);
        }
      });
      nuxt.hook("prepare:types", ({ references }) => {
        references.push({ path: resolve(nuxt.options.buildDir, `module/${moduleName}-schema.d.ts`) });
      });
    }
  }
});

export { module as default };
